@@ -19,7 +19,12 @@
 
 #include "kerncompat.h"
 #include <sys/ioctl.h>
+#include <sys/time.h>
 #include <sys/sysinfo.h>
+#include <sys/select.h>
+#include <sys/stat.h>
+#include <dirent.h>
+#include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -28,15 +33,19 @@
 #include <mntent.h>
 #include <ctype.h>
 #include <limits.h>
-#include <uuid/uuid.h>
-#include "kernel-shared/uapi/btrfs.h"
+#include <strings.h>
+#include "kernel-lib/list.h"
+#include "kernel-shared/accessors.h"
 #include "kernel-shared/ctree.h"
 #include "kernel-shared/disk-io.h"
 #include "kernel-shared/volumes.h"
 #include "common/utils.h"
+#include "common/device-utils.h"
 #include "common/path-utils.h"
 #include "common/open-utils.h"
+#include "common/sysfs-utils.h"
 #include "common/messages.h"
+#include "common/tree-search.h"
 #include "cmds/commands.h"
 #include "mkfs/common.h"
 
@@ -66,18 +75,6 @@ void btrfs_format_csum(u16 csum_type, const u8 *data, char *output)
 	}
 }
 
-int get_device_info(int fd, u64 devid,
-		struct btrfs_ioctl_dev_info_args *di_args)
-{
-	int ret;
-
-	di_args->devid = devid;
-	memset(&di_args->uuid, '\0', sizeof(di_args->uuid));
-
-	ret = ioctl(fd, BTRFS_IOC_DEV_INFO, di_args);
-	return ret < 0 ? -errno : 0;
-}
-
 int get_df(int fd, struct btrfs_ioctl_space_args **sargs_ret)
 {
 	u64 count = 0;
@@ -124,11 +121,10 @@ int get_df(int fd, struct btrfs_ioctl_space_args **sargs_ret)
 }
 
 
-static u64 find_max_device_id(struct btrfs_ioctl_search_args *search_args,
-			      int nr_items)
+static u64 find_max_device_id(struct btrfs_tree_search_args *args, int nr_items)
 {
 	struct btrfs_dev_item *dev_item;
-	char *buf = search_args->buf;
+	char *buf = btrfs_tree_search_data(args, 0);
 
 	buf += (nr_items - 1) * (sizeof(struct btrfs_ioctl_search_header)
 				       + sizeof(struct btrfs_dev_item));
@@ -145,8 +141,8 @@ static int search_chunk_tree_for_fs_info(int fd,
 	int ret;
 	int max_items;
 	u64 start_devid = 1;
-	struct btrfs_ioctl_search_args search_args;
-	struct btrfs_ioctl_search_key *search_key = &search_args.key;
+	struct btrfs_tree_search_args args;
+	struct btrfs_ioctl_search_key *sk;
 
 	fi_args->num_devices = 0;
 
@@ -154,41 +150,41 @@ static int search_chunk_tree_for_fs_info(int fd,
 	       / (sizeof(struct btrfs_ioctl_search_header)
 			       + sizeof(struct btrfs_dev_item));
 
-	search_key->tree_id = BTRFS_CHUNK_TREE_OBJECTID;
-	search_key->min_objectid = BTRFS_DEV_ITEMS_OBJECTID;
-	search_key->max_objectid = BTRFS_DEV_ITEMS_OBJECTID;
-	search_key->min_type = BTRFS_DEV_ITEM_KEY;
-	search_key->max_type = BTRFS_DEV_ITEM_KEY;
-	search_key->min_transid = 0;
-	search_key->max_transid = (u64)-1;
-	search_key->nr_items = max_items;
-	search_key->max_offset = (u64)-1;
+	memset(&args, 0, sizeof(args));
+	sk = btrfs_tree_search_sk(&args);
+	sk->tree_id = BTRFS_CHUNK_TREE_OBJECTID;
+	sk->min_objectid = BTRFS_DEV_ITEMS_OBJECTID;
+	sk->min_type = BTRFS_DEV_ITEM_KEY;
+	sk->max_objectid = BTRFS_DEV_ITEMS_OBJECTID;
+	sk->max_type = BTRFS_DEV_ITEM_KEY;
+	sk->min_transid = 0;
+	sk->max_transid = (u64)-1;
+	sk->nr_items = max_items;
+	sk->max_offset = (u64)-1;
 
 again:
-	search_key->min_offset = start_devid;
+	sk->min_offset = start_devid;
 
-	ret = ioctl(fd, BTRFS_IOC_TREE_SEARCH, &search_args);
+	ret = btrfs_tree_search_ioctl(fd, &args);
 	if (ret < 0)
 		return -errno;
 
-	fi_args->num_devices += (u64)search_key->nr_items;
+	fi_args->num_devices += (u64)sk->nr_items;
 
-	if (search_key->nr_items == max_items) {
-		start_devid = find_max_device_id(&search_args,
-					search_key->nr_items) + 1;
+	if (sk->nr_items == max_items) {
+		start_devid = find_max_device_id(&args, sk->nr_items) + 1;
 		goto again;
 	}
 
 	/* Get the latest max_id to stay consistent with the num_devices */
-	if (search_key->nr_items == 0)
+	if (sk->nr_items == 0)
 		/*
 		 * last tree_search returns an empty buf, use the devid of
 		 * the last dev_item of the previous tree_search
 		 */
 		fi_args->max_id = start_devid - 1;
 	else
-		fi_args->max_id = find_max_device_id(&search_args,
-						search_key->nr_items);
+		fi_args->max_id = find_max_device_id(&args, sk->nr_items);
 
 	return 0;
 }
@@ -215,7 +211,6 @@ int get_fs_info(const char *path, struct btrfs_ioctl_fs_info_args *fi_args,
 	struct btrfs_ioctl_dev_info_args *di_args;
 	struct btrfs_ioctl_dev_info_args tmp;
 	char mp[PATH_MAX];
-	DIR *dirstream = NULL;
 
 	memset(fi_args, 0, sizeof(*fi_args));
 
@@ -255,9 +250,9 @@ int get_fs_info(const char *path, struct btrfs_ioctl_fs_info_args *fi_args,
 	}
 
 	/* at this point path must not be for a block device */
-	fd = open_file_or_dir(path, &dirstream);
+	fd = btrfs_open_file_or_dir(path);
 	if (fd < 0) {
-		ret = -errno;
+		ret = fd;
 		goto out;
 	}
 
@@ -280,7 +275,7 @@ int get_fs_info(const char *path, struct btrfs_ioctl_fs_info_args *fi_args,
 		 * search_chunk_tree_for_fs_info() will lacks the devid 0
 		 * so manual probe for it here.
 		 */
-		ret = get_device_info(fd, 0, &tmp);
+		ret = device_get_info(fd, 0, &tmp);
 		if (!ret) {
 			fi_args->num_devices++;
 			ndevs++;
@@ -303,7 +298,7 @@ int get_fs_info(const char *path, struct btrfs_ioctl_fs_info_args *fi_args,
 		memcpy(di_args, &tmp, sizeof(tmp));
 	for (; last_devid <= fi_args->max_id && ndevs < fi_args->num_devices;
 	     last_devid++) {
-		ret = get_device_info(fd, last_devid, &di_args[ndevs]);
+		ret = device_get_info(fd, last_devid, &di_args[ndevs]);
 		if (ret == -ENODEV)
 			continue;
 		if (ret)
@@ -321,7 +316,7 @@ int get_fs_info(const char *path, struct btrfs_ioctl_fs_info_args *fi_args,
 	}
 
 out:
-	close_file_or_dir(fd, dirstream);
+	close(fd);
 	return ret;
 }
 
@@ -342,8 +337,23 @@ int get_fsid(const char *path, u8 *fsid, int silent)
 {
 	int ret;
 	int fd;
+	int flags = O_RDONLY;
+	struct stat st;
 
-	fd = open(path, O_RDONLY);
+	ret = stat(path, &st);
+	if (ret < 0) {
+		if (!silent)
+			error("failed to stat %s: %m", path);
+		return -errno;
+	}
+	/*
+	 * Open in non-blocking mode in case that path is a fifo or a special
+	 * character device where opening gets stuck (but is interruptible).
+	 */
+	if ((st.st_mode & S_IFMT) == S_IFCHR || (st.st_mode & S_IFMT) == S_IFIFO)
+		flags |= O_NONBLOCK;
+
+	fd = open(path, flags);
 	if (fd < 0) {
 		if (!silent)
 			error("failed to open %s: %m", path);
@@ -406,6 +416,7 @@ int ask_user(const char *question)
 	char *answer;
 
 	printf("%s [y/N]: ", question);
+	fflush(stdout);
 
 	return fgets(buf, sizeof(buf) - 1, stdin) &&
 	       (answer = strtok_r(buf, " \t\n\r", &saveptr)) &&
@@ -945,10 +956,30 @@ u8 rand_u8(void)
 	return (u8)(rand_u32());
 }
 
+/*
+ * Parse a boolean value from an environment variable.
+ *
+ * As long as the environment variable is not set to "0", "n" or "\0",
+ * it would return true.
+ */
+bool get_env_bool(const char *env_name)
+{
+	char *env_value_str;
+
+	env_value_str = getenv(env_name);
+	if (!env_value_str)
+		return false;
+	if (env_value_str[0] == '0' || env_value_str[0] == 'n' ||
+	    env_value_str[0] == 0)
+		return false;
+	return true;
+}
+
 void btrfs_config_init(void)
 {
 	bconf.output_format = CMD_FORMAT_TEXT;
 	bconf.verbose = BTRFS_BCONF_UNSET;
+	INIT_LIST_HEAD(&bconf.params);
 }
 
 void bconf_be_verbose(void)
@@ -964,6 +995,57 @@ void bconf_be_quiet(void)
 	bconf.verbose = BTRFS_BCONF_QUIET;
 }
 
+void bconf_add_param(const char *key, const char *value)
+{
+	struct config_param *param;
+
+	param = calloc(1, sizeof(*param));
+	if (!param)
+		return;
+	param->key = strdup(key);
+	if (value)
+		param->value = strdup(value);
+	list_add(&param->list, &bconf.params);
+}
+
+const char *bconf_param_value(const char *key)
+{
+	struct config_param *param;
+
+	list_for_each_entry(param, &bconf.params, list) {
+		if (strcmp(key, param->key) == 0)
+			return param->value;
+	}
+	return NULL;
+}
+
+void bconf_save_param(const char *str)
+{
+	char *tmp;
+
+	tmp = strchr(str, '=');
+	if (!tmp) {
+		bconf_add_param(str, NULL);
+		printf("Global param: %s\n", str);
+	} else {
+		*tmp = 0;
+		bconf_add_param(str, tmp + 1);
+		printf("Global param: %s=%s\n", str, tmp + 1);
+		*tmp = '=';
+	}
+}
+
+void bconf_set_dry_run(void)
+{
+	pr_verbose(LOG_INFO, "Dry-run requested\n");
+	bconf.dry_run = 1;
+}
+
+bool bconf_is_dry_run(void)
+{
+	return bconf.dry_run == 1;
+}
+
 /* Returns total size of main memory in bytes, -1UL if error. */
 unsigned long total_memory(void)
 {
@@ -1115,7 +1197,11 @@ out:
  */
 char *btrfs_test_for_multiple_profiles(int fd)
 {
-	char *data, *metadata, *system, *mixed, *types;
+	char *data = NULL;
+	char *metadata = NULL;
+	char *system = NULL;
+	char *mixed = NULL;
+	char *types = NULL;
 
 	btrfs_get_string_for_multiple_profiles(fd, &data, &metadata, &mixed,
 			&system, &types);
@@ -1160,78 +1246,6 @@ void btrfs_warn_experimental(const char *str)
 #endif
 }
 
-/*
- * Open a file in fsid directory in sysfs and return the file descriptor or
- * error
- */
-int sysfs_open_fsid_file(int fd, const char *filename)
-{
-	u8 fsid[BTRFS_UUID_SIZE];
-	char fsid_str[BTRFS_UUID_UNPARSED_SIZE];
-	char sysfs_file[PATH_MAX];
-	int ret;
-
-	ret = get_fsid_fd(fd, fsid);
-	if (ret < 0)
-		return ret;
-	uuid_unparse(fsid, fsid_str);
-
-	ret = path_cat3_out(sysfs_file, "/sys/fs/btrfs", fsid_str, filename);
-	if (ret < 0)
-		return ret;
-
-	return open(sysfs_file, O_RDONLY);
-}
-
-/*
- * Open a file in the toplevel sysfs directory and return the file descriptor
- * or error.
- */
-int sysfs_open_file(const char *name)
-{
-	char path[PATH_MAX];
-	int ret;
-
-	ret = path_cat_out(path, "/sys/fs/btrfs", name);
-	if (ret < 0)
-		return ret;
-	return open(path, O_RDONLY);
-}
-
-/*
- * Open a directory by name in fsid directory in sysfs and return the file
- * descriptor or error, filedescriptor suitable for fdreaddir. The @dirname
- * must be a directory name.
- */
-int sysfs_open_fsid_dir(int fd, const char *dirname)
-{
-	u8 fsid[BTRFS_UUID_SIZE];
-	char fsid_str[BTRFS_UUID_UNPARSED_SIZE];
-	char sysfs_file[PATH_MAX];
-	int ret;
-
-	ret = get_fsid_fd(fd, fsid);
-	if (ret < 0)
-		return ret;
-	uuid_unparse(fsid, fsid_str);
-
-	ret = path_cat3_out(sysfs_file, "/sys/fs/btrfs", fsid_str, dirname);
-	if (ret < 0)
-		return ret;
-
-	return open(sysfs_file, O_DIRECTORY | O_RDONLY);
-}
-
-/*
- * Read up to @size bytes to @buf from @fd
- */
-int sysfs_read_file(int fd, char *buf, size_t size)
-{
-	lseek(fd, 0, SEEK_SET);
-	memset(buf, 0, size);
-	return read(fd, buf, size);
-}
-
 static const char exclop_def[][16] = {
 	[BTRFS_EXCLOP_NONE]		= "none",
 	[BTRFS_EXCLOP_BALANCE]		= "balance",
@@ -1302,9 +1316,9 @@ int check_running_fs_exclop(int fd, enum exclusive_operation start, bool enqueue
 
 	sysfs_fd = sysfs_open_fsid_file(fd, "exclusive_operation");
 	if (sysfs_fd < 0) {
-		if (errno == ENOENT)
+		if (sysfs_fd == -ENOENT)
 			return 0;
-		return -errno;
+		return sysfs_fd;
 	}
 
 	exclop = get_fs_exclop(fd);
@@ -1329,34 +1343,58 @@ int check_running_fs_exclop(int fd, enum exclusive_operation start, bool enqueue
 			get_fs_exclop_name(exclop));
 		ret = 1;
 		goto out;
+	} else {
+		pr_verbose(LOG_DEFAULT, "Waiting for another exclusive operation '%s' to finish ...",
+			get_fs_exclop_name(exclop));
+		fflush(stdout);
 	}
 
+	/*
+	 * The sysfs file descriptor needs to be reopened and all data read
+	 * before each select().
+	 */
 	while (exclop > 0) {
 		fd_set fds;
 		struct timeval tv = { .tv_sec = 60, .tv_usec = 0 };
+		char tmp[1024];
 
+		close(sysfs_fd);
+		sysfs_fd = sysfs_open_fsid_file(fd, "exclusive_operation");
+		if (sysfs_fd < 0)
+			return sysfs_fd;
 		FD_ZERO(&fds);
 		FD_SET(sysfs_fd, &fds);
 
+		ret = read(sysfs_fd, tmp, sizeof(tmp));
 		ret = select(sysfs_fd + 1, NULL, NULL, &fds, &tv);
 		if (ret < 0) {
 			ret = -errno;
 			break;
 		}
 		if (ret > 0) {
+			close(sysfs_fd);
+			sysfs_fd = sysfs_open_fsid_file(fd, "exclusive_operation");
+			if (sysfs_fd < 0)
+				return sysfs_fd;
+
+			FD_ZERO(&fds);
+			FD_SET(sysfs_fd, &fds);
+
+			ret = read(sysfs_fd, tmp, sizeof(tmp));
 			/*
 			 * Notified before the timeout, check again before
 			 * returning. In case there are more operations
 			 * waiting, we want to reduce the chances to race so
 			 * reuse the remaining time to randomize the order.
 			 */
-			tv.tv_sec /= 2;
+			tv.tv_sec = (tv.tv_sec % 10) + 1;
 			ret = select(sysfs_fd + 1, NULL, NULL, &fds, &tv);
 			exclop = get_fs_exclop(fd);
 			if (exclop <= 0)
 				ret = 0;
 		}
 	}
+	pr_verbose(LOG_DEFAULT, " done\n");
 out:
 	close(sysfs_fd);
 
