@@ -422,6 +422,10 @@ int main(int argc, char **argv) {
 	else
 		flag = IO;
 	progress_init(&prog, start, stop, fs_info.totalblock, flag, fs_info.block_size);
+        prog.prog_second = opt.prog_second;
+        prog.binary_prefix = opt.binary_prefix;
+        if (opt.prog_second)
+            strncpy(prog.time_unit, "sec", 4);
 	copied = 0;				/// initial number is 0
 
 	/**
